module gauleg_module

implicit none

public :: gauleg

intrinsic abs, cos, atan 

integer, public, parameter :: real8 = selected_real_kind(10)

contains
 
!
!***********************************************************************
!  this subroutine is used by the program GaussLegendre.f90
!***********************************************************************
!
      subroutine gauleg(x1,x2,x,w,n,eps)
      real(kind=real8), intent(in)                  :: x1
      real(kind=real8), intent(in)                  :: x2
      real(kind=real8), intent(inout), dimension(:) :: x  !(n)
      real(kind=real8), intent(inout), dimension(:) :: w  !(n)
      real(kind=real8), intent(in)                  :: eps
      integer         , intent(in)                  :: n
!
!***********************************************************************
!     given the lower and upper limits of integration x1 and x2,and
!     given n, this subroutine returns arrays x and w of length n,
!     containing the abscissae and weights of the gauss-legendre n-point
!     quadrature formula. the algorithm is given in "numerical recipes"
!     pages 125-126 and is rewritten here into double precision.
!***********************************************************************
!
!***********************************************************************
!     converted f90, 200409. m. r. schaferkotter. schaferk@bellsouth.net
!***********************************************************************
!
      real(kind=real8) :: xm,xl,z,p1,p2,p3,pp,z1,pi
      integer          :: i,j,m
!
!
      pi = (4.0_real8)*atan(1.0_real8)
      m  = (n+1)/2
      xm = 0.5_real8*(x2+x1)
      xl = 0.5_real8*(x2-x1)
!
      do i = 1,m
        z = cos(pi*(i-0.25_real8)/(n+0.5_real8))
!
!***********************************************************************
!     starting with the above approximation to the i-th root, we enter
!     the main loop of refinement by newton's method.
!***********************************************************************
!
1     continue
!
        p1 = 1.0_real8
        p2 = 0.0_real8
!
        do j = 1,n
!
!***********************************************************************
!     loop up the recurrence relation to get the legendre polynomial
!     evaluated at z.
!***********************************************************************
!
          p3 = p2
          p2 = p1
          p1 = ((2.0_real8*j-1.0_real8)*z*p2-(j-1.0_real8)*p3)/j

        enddo
!
!***********************************************************************
!     p1 is now the desired legendre polynomial. we next compute pp,
!     its derivative, by a standard relation involving also p2, the
!     polynomial of one lower order.
!***********************************************************************
!
        pp = n*(z*p1-p2)/(z*z-1.0_real8)
        z1 = z
        z  = z1-p1/pp
!
        if (abs(z-z1) > eps) then
          go to 1
        endif
!
        x((n/2)+1-i) = xm+xl*z
        w((n/2)+1-i) = 2.0d0*xl/((1.0d0-z*z)*pp*pp)
!
      enddo

   endsubroutine gauleg 

end module gauleg_module

