module fft_module

implicit none

public :: fft

intrinsic real

integer, public, parameter :: real8 = selected_real_kind(10)           !double precision
integer, public, parameter :: real4 = selected_real_kind( 6,37)        !single precision

integer, public, parameter :: rl    = real4

contains
 
      subroutine fft(fr,fi,mp,isw)
      real(kind=rl), intent(inout), dimension(:) :: fr
      real(kind=rl), intent(inout), dimension(:) :: fi
      integer      , intent(in)                  :: mp
      integer      , intent(in)                  :: isw 
!
!***********************************************************************
!  compute the discrete Fourier transform
!
!    ft(k+1) = scale*sum f(j+1)*exp(isn*2*pi*i*j*k/n), k=0,1,...,n-1
!
!      of the complex array
!
!      f(j+1)=fr(j+1)+i*fi(j+1), j=0,1, ..., n-1, where
!
!      n = 2**mp
!      i = sqrt(-1)
!      isn = +1 if isw > 0
!          = -1 if isw < 0
!    scale = 1       	if isw=-1 or +1		
!          =1/n		if isw=-2 or +2
!          =1/sqrt(n)	if isw=-3 or +3
!
!  computations are done in place, 
!  upon return the real arrays fr, fi have been overwritten with the real and imaginary parts
!  of the Fourier transform.
!
!  The code is based on the presentation of the FFT given in the text
!  'A First Course in Fourier Analysis by David W. Kammler,
!
!  converted to f90, 200409, m. r. schaferkotter. schaferk@bellsouth.net
!
!***********************************************************************
!
      real(kind=rl), parameter :: zero   = 0.0
      real(kind=rl), parameter :: one    = 1.0
      real(kind=rl), parameter :: two    = 2.0
!
      integer      , parameter :: maxmp  = 10
      integer      , parameter :: maxds  = 257
      integer      , parameter :: maxdir = 32
!
!***********************************************************************
!  for larger vectors f, change the value of maxmp and
!    the corresponding values of maxds=2**(maxmp-2)+1 and
!    maxdir=2**[(maxmp+1)/2] in the above parameter statement.
!***********************************************************************
!
      real(kind=rl), dimension(maxds)  :: s
      real(kind=rl), dimension(maxdir) :: ir
!
      integer       :: i
      integer       :: indx
      integer       :: ip 
      integer       :: iq
      integer       :: irp1
      integer       :: irpp
      integer       :: isws
      integer       :: it
      integer       :: itemp
      integer       :: j
      integer       :: k
      integer       :: k0
      integer       :: k1
      integer       :: k2
      integer       :: k3
      integer       :: k4m
      integer       :: kcap
      integer       :: kt2
      integer       :: lambda 
      integer, save :: lastmp = 0
      integer       :: m
      integer       :: maxs
      integer       :: mcap
      integer       :: mu
      integer       :: muplus
      integer       :: n
      integer       :: ne
      integer       :: nh
      integer       :: npr
      integer       :: npr1
      integer       :: nq
      integer       :: nsn
      integer       :: nu
    
      real(kind=rl) :: cs 
      real(kind=rl) :: h 
      real(kind=rl) :: r1
      real(kind=rl) :: r2
      real(kind=rl) :: rn
      real(kind=rl) :: rroot2
      real(kind=rl) :: rrootm
      real(kind=rl) :: rrootn
      real(kind=rl) :: scale 
      real(kind=rl) :: sn
      real(kind=rl) :: temp 
      real(kind=rl) :: temp1 
      real(kind=rl) :: temp2
      real(kind=rl) :: tempi
      real(kind=rl) :: tempr 
!
      maxs   = 0
      isws   = isw**2
!
      if((mp < 0) .or. (mp > maxmp)) then
        write(unit=*,fmt=*) "fft: Improper argument mp < 0 .or. mp > maxmp."," mp= ",mp," maxmp= ", maxmp
  	stop
      endif

      if((isws == 0) .or. (isws > 9)) then
        write(unit=*,fmt=*) "fft: Improper argument isws ==0 .or. isws > 9.", " isws=",isws
  	stop
      endif
!
      if(mp == 0) return
!
      if(mp == lastmp) then
        if(mp == 1) go to 20
        if(mp > 1) go to 14
      endif
!
!***********************************************************************
!	At the time of the first call of the subroutine with a given value of
!	mp > 0, initialize various constants and arrays that depend on mp
!	but not on fr,fi.  On subsequent calls with the same mp, bypass this
!	initialization process.
!***********************************************************************
!
      lastmp = mp
      n      = 2**mp
      temp   = n
      rn     = one/temp
      rrootn = sqrt(rn)
      nh     = n/2
!
      if(mp == 1) go to 20
      if(mp == 2) go to 8
!
      if(mp < maxs) then
        nsn = 2**(maxs-mp)
      else
        nsn = 1
      endif
!
!***********************************************************************
!	When mp <= maxs, the spacing parameter is used to retrieve sine values
!	from a previously computed table.
!***********************************************************************
!
      if(mp <= maxs) go to 8
!
!***********************************************************************
!	When n=8,16,32, ... and mp > maxs, precompute
!	s(j+1)=sin((pi/2)*(j/nq)), j=0,1,2,...,nq, for use in
!	the subsequent calculations.
!***********************************************************************
!
      nq      = n/4
      ne      = n/8
      rroot2  = one/sqrt(two)
      maxs    = mp
      s(1)    = zero          !=sin(0*pi/4)
      s(ne+1) = rroot2        !=sin(1*pi/4)
      s(nq+1) = one           !=sin(2*pi/4)

      if(mp == 3) go to 8
!
      h = rroot2              !=0.5*sec(pi/4)
!
!***********************************************************************
!  Pass from the course grid to a finer one using the trig identity:
!  sin(s)=(0.5*sec(b))*(sin(a-b)+sin(a+b)).
!
!  (O. Bunemann, cf. SIAM J. SCI. STAT. COMPUT. 7 (1986), pp. 624-638.)
!***********************************************************************
!
      k = ne
      do i=4,mp
        h   = one/sqrt(two+one/h) !=half secant of half the previous angle
        kt2 = k
        k   = k/2                 !=n/2**i

        do j=k,nq,kt2
          s(j+1) = h*(s(j-k+1)+s(j+k+1))
        enddo
!
      enddo
!
 8	continue
!
!***********************************************************************
!	Prepare a short table of bit reversed integers to use in the
!	subsequent bit reversal permutation.
!***********************************************************************
!
	muplus = (mp+1)/2
	m      = 1
	ir(1)  = 0
!
	do nu=1,muplus
!
	  do k=1,m
            it      = 2*ir(k)
            ir(k)   = it
            ir(k+m) = it+1
          enddo
!
 	  m = m+m
!
        enddo
!
!***********************************************************************
! 	If mp is odd, then m=m/2.
!***********************************************************************
!
        itemp = 2*(mp/2)
!
        if(itemp /= mp) then
          m = m/2
        endif
!	
!***********************************************************************
!  The parameters n,nh,nq,ne,m,lastmp,maxs,nsn,rn,rrootn,rroot2 and the
!  arrays ir(*), s(*) are now suitably initialized.
!***********************************************************************
!
 14	continue
!
!***********************************************************************
!	Apply the bit reversal permutation to the array f using
!	Bracewell-Buneman scheme.
!***********************************************************************
!
	do iq=1,m-1
!
	  npr  = iq-m
	  irpp = ir(iq+1)*m 
!
	  do ip=0,ir(iq+1)-1
            npr      = npr+m
            npr1     = npr+1
            irp1     = irpp+ir(ip+1)+1
            tempr    = fr(npr1)
            tempi    = fi(npr1)
            fr(npr1) = fr(irp1)
            fi(npr1) = fi(irp1)
            fr(irp1) = tempr
            fi(irp1) = tempi
          enddo
        enddo
!
 20		continue
!
!***********************************************************************
!	Apply the mp Q-matrices.
!
!	Carry out stage 1 of the FFT using blocks of size 2x2 and apply
!	the desired scale factor.
!***********************************************************************
!
	if(isws == 1) then
	  	scale = one
	  elseif(isws == 4) then
	  	scale = rn
	  else
	  	scale = rrootn
	  endif
!
	 do k=0,nh-1
	   k1 = 2*k+1
	   k2 = k1+1
	   tempr = (fr(k1)-fr(k2))*scale
	   tempi = (fi(k1)-fi(k2))*scale
	   fr(k1) = (fr(k1)+fr(k2))*scale
	   fi(k1) = (fi(k1)+fi(k2))*scale
	   fr(k2) = tempr
 	   fi(k2) = tempi
         enddo
!
         if(mp == 1) return
!
!***********************************************************************
!	Carry out stages 2,3,...,mp of the FFT using blocks of size mxm=4x4,8x8,...,nxn.
!***********************************************************************
!
	mcap = 1
	kcap = n/4
!
	do 32 mu=2,mp
!
!***********************************************************************
!	At this point mcap=2**(mu-2) and kcap=2**(mp-mu).
!***********************************************************************
!
!
!***********************************************************************
!     Deal first with the quadruplet of components where sin=0 or cos=0.
!***********************************************************************
!
      do k=0,kcap-1
        k0     = k*4*mcap+1
        k1     = k0+mcap
        k2     = k0+2*mcap
        k3     = k0+3*mcap
        tempr  = fr(k0)-fr(k2)
        tempi  = fi(k0)-fi(k2)
        fr(k0) = fr(k0)+fr(k2)
        fi(k0) = fi(k0)+fi(k2)
        fr(k2) = tempr
        fi(k2) = tempi
!
        if(isw < 0) then
          fr(k3) = -fr(k3)
          fi(k3) = -fi(k3)
        endif
!
        temp1 = fr(k1)+fi(k3)
        temp2 = fi(k1)-fr(k3)
        fr(k1) = fr(k1)-fi(k3)
        fi(k1) = fi(k1)+fr(k3)
        fr(k3) = temp1
        fi(k3) = temp2
      enddo
!
      if(mcap == 1) go to 30
!
!***********************************************************************
!  the remaining mcap-1 quadruplets of components where sin, cos are both nonzero.
!***********************************************************************
!
       do lambda=1,mcap-1
         indx = nsn*lambda*kcap+1
         sn   = s(indx)                    !=sin((2*pi)*(lambda/mcap))
         indx = nq-indx+2
         cs   = s(indx)                    !=cos((2*pi)*(lambda/mcap))

         if(isw < 0) then
           sn = -sn
	 endif
!
         do k=0,kcap-1
           k4m    = k*4*mcap+1
           k0     = k4m+lambda
           k1     = k4m+2*mcap-lambda
           k2     = k4m+2*mcap+lambda  
           k3     = k4m+4*mcap-lambda
           r1     = cs*fr(k2)-sn*fi(k2)
           r2     = cs*fi(k2)+sn*fr(k2)
           temp1  = fr(k0)-r1
           temp2  = fi(k0)-r2
           fr(k0) = fr(k0)+r1
           fi(k0) = fi(k0)+r2
           fr(k2) = temp1
           fi(k2) = temp2
           r1     = cs*fr(k3)+sn*fi(k3)
           r2     = cs*fi(k3)-sn*fr(k3)
           temp1  = fr(k1)+r1
           temp2  = fi(k1)+r2
           fr(k1) = fr(k1)-r1
           fi(k1) = fi(k1)-r2
           fr(k3) = temp1
           fi(k3) = temp2
         enddo
       enddo 
 30    mcap = mcap*2
       kcap = kcap/2
 32    continue
 
   endsubroutine fft 
!
end module fft_module
!
