module hartley2d_module
implicit none
!
!***********************************************************************
! obtained from http://users.bigpond.net.au/amiller/
!  hartly2d.f90   Hartley 2D Fast Fourier Transform.   
!
!  reformatted and presented in F language style m.r. schaferkotter 200410
!***********************************************************************
!
!
!***********************************************************************
!! From hamill@mailhub.nmg.sms.siemens.com Wed Apr 24 16:21:31 1996
!
!! Code converted using TO_F90 by Alan Miller
!! Date: 2004-02-04  Time: 13:01:18
!
!! I see from the newsgroups that you are interested in multidimensional
!! Hartley transforms.  The attached FORTRAN code works (I think ... but
!! it's been a few years since I used it) for 2D and 1D.  For 3D see H.Hao
!! and R.N.Bracewell, "A Three-Dimensional DFT Algorithm Using the Fast
!! Hartley Transform," Proc IEEE vol. 75 no. 2 (Feb, 87) p.264-266.
!
!! I do not vouch for the correctness of the information or the code, but
!! perhaps it is a useful starting point.
!  
!***********************************************************************
!
!
!***********************************************************************
!  this subprogram is used by the program xfht.f90
!***********************************************************************
!
public :: hartley2d

private :: fhtfr

intrinsic int, log, real

integer, private, parameter :: real8 = selected_real_kind(10)

contains
 
      subroutine hartley2d(nx, f2)
      integer, intent(in)                              :: nx
      real   , intent(inout), dimension(0:nx-1,0:nx-1) :: f2
!
!***********************************************************************
!  a subroutine for the 2-dimensional hartley transform of an 
!    array f2(nx,nx),  where nx = 2**p
!  this routine calls the 1-d hartley transform routine fhtfr, for rows, 
!    followed by columns.
!
! for a 3d transform,  
!  H.Hao and R.N.Bracewell, "A three-dimensional dft algorithm using the fast hartley transform.",
!   Proc Ieee vol. 75 no. 2 p.264-266.
!***********************************************************************
!
      real                      :: a, b, c, d, e
      real   , dimension(0:255) :: f
      integer                   :: i, j, n, ny, n1, p, u, v

      p = int( log(real(nx))/log(2.0) + 0.5)
      n  = 2**p
      n1 = n - 1

      do j = 0,n1
        call fhtfr(p, f2(0,j))
      end do

      do i = 0,n1

        do j = 0,n1
          f(j)=f2(i,j)
        end do

        call fhtfr(p, f)

        do j = 0,n1
          f2(i,j)=f(j)
        end do

     end do
!
!***********************************************************************
!  up to this point f2(u,v) now contain the separable
!    cas(2*pi*u*x)*cas(2*pi*v*y)   transform
!  where 'cas' is  cos(x)+sin(x)
!***********************************************************************
!
!
!***********************************************************************
!  the following replacements form the nonseparable 2d-hartley transform
!    ref:   
! r.n. bracewell, et al., "fast two-dimensional hartley transform", proc ieee, 74:9,1282 
!***********************************************************************
!
      ny = nx

      do v = 1,ny/2-1
        do u = 1,nx/2-1

          a = f2(u   ,v)
          b = f2(nx-u,v)
          c = f2(u   ,ny-v)
          d = f2(nx-u,ny-v)
          e = 0.5*( (a+d)-(b+c) )

          f2(u,v)       = a - e
          f2(nx-u,v)    = b + e
          f2(u,ny-v)    = c + e
          f2(nx-u,ny-v) = d - e

        end do
      end do
 
   endsubroutine hartley2d 

 
      subroutine fhtfr(p, f)
      integer, intent(in)                      :: p
      real   , intent(inout), dimension(0:255) :: f

!
!***********************************************************************
!   subroutine fhtfor takes input f() and returns the dht in the same f()
!   length of f() is 2**p
!
!     this is the corrected version of fhtfor.for
!     rn bracewell,  the hartley transform, oxford univ. press (1986) pp. 140-144
!***********************************************************************
!
      integer                  :: d9
      integer                  :: e9
      integer                  :: i
      integer                  :: j
      integer                  :: k9
      integer                  :: l9
      integer                  :: n
      integer                  :: n9
      integer                  :: q9
      integer                  :: sx9
      integer                  :: u9
      integer                  :: x9
      integer                  :: y9
      integer, dimension(0:64) :: a9
      integer, dimension(0:10) :: m9
      integer, dimension(0:10) :: c9
      integer, dimension(0:10) :: v9

      real   , dimension(256)  :: s9
      real   , dimension(256)  :: t9

      real                     :: h9
      real                     :: t
!
      real   , parameter       :: pi   = 3.1415926536_real8
      integer                  :: pold = -99
!
!
      if (p == 1) then
        j    = f(0) + f(1)
        f(1) = f(0) - f(1)
        f(0) = j
        go to 9636   !   return
      end if
!
      if(p == pold) go to 9400
!
      n9 = 2**(p-2)
      n = 4*n9
      c9(5) = n-1
      c9(6) = p-1
!
!***********************************************************************
!   get powers of 2
!***********************************************************************
!
      i = 1
      m9(0) = 1
      m9(1) = 2

9204   m9(i+1) = m9(i) + m9(i)

      i = i + 1

      if(i <= p) go to 9204
!
!***********************************************************************
!  trigonometric function tables
!***********************************************************************
!
      if(n == 4) go to 9400

      if(n == 8) then

        s9(2) = 1.
        s9(1) = sin(pi/4.0)

        go to 9330
      end if
!
!***********************************************************************
!   get sines
!***********************************************************************
!
      s9(n9) = 1.0
!
!***********************************************************************
!   coarse seed table for sines
!***********************************************************************
!
      do i = 1,3
        s9(i*n9/4) = sin(i*pi/8.0)
      end do
!
!***********************************************************************
!   initial half secant
!***********************************************************************
!
      h9 = 1.0/2.0/cos(pi/16.0)
!
!***********************************************************************
!   fill sine table
!***********************************************************************
!
      c9(4) = p-4

      do i = 1,p-4
        c9(4) = c9(4)-1
        v9(0) = 0

        do j = m9(c9(4)), n9-m9(c9(4)), m9(c9(4)+1)
          v9(1) = j + m9(c9(4))
          s9(j) = h9*(s9(v9(1))+v9(0))
          v9(0) = s9(v9(1))
        end do
!
!***********************************************************************
!   half secant recursion
!***********************************************************************
!
        h9 = 1/sqrt(2+1/h9)
      end do
!
!***********************************************************************
!   get tangents
!***********************************************************************
!
9330 c9(0) = n9 - 1

      do i = 1,n9-1
        t9(i) = (1-s9(c9(0)))/s9(i)
        c9(0) = c9(0) - 1
      end do
!
      t9(n9) = 1
!
!***********************************************************************
!   fast permute
!   for p = 2,3 permute directly
!***********************************************************************
!
9400 if(p == 2) then
       v9(9) = f(1)
       f(1)  = f(2)
       f(2)  = v9(9)
       go to 9500
     end if
!
     if(p == 3) then
       v9(9) = f(1)
       f(1)  = f(4)
       f(4)  = v9(9)
       v9(9) = f(3)
       f(3)  = f(6)
       f(6)  = v9(9)
       go to 9500
     end if
!
     if(p == pold) go to 9420
!
!***********************************************************************
!   for p=4, 5, 6 (q9=2,3), skip structure table
!***********************************************************************
!
     q9 = p/2
     c9(2) = m9(q9)
     q9 = q9 + mod(p,2)
!
     if(q9 == 2) then
       a9(0) = 0
       a9(1) = 2
       a9(2) = 1
       a9(3) = 3
       go to 9420
     end if
!
      if(q9 == 3) then
        a9(0) = 0
        a9(1) = 4
        a9(2) = 2
        a9(3) = 6
        a9(4) = 1
        a9(5) = 5
        a9(6) = 3
        a9(7) = 7
        go to 9420
      end if
!
!***********************************************************************
!   set up structure table
!***********************************************************************
!
      a9(0) = 0
      a9(1) = 1
!
      do i = 2,q9
        do j = 0,m9(i-1)-1
          a9(j) = a9(j) + a9(j)
          a9(j+m9(i-1))=a9(j)+1
        enddo
      enddo
!
!***********************************************************************
!   permute
!***********************************************************************
!
9420  do i = 1,c9(2)-1
        v9(4)  = c9(2)*a9(i)
        v9(5)  = i
        v9(6)  = v9(4)
        v9(7)  = f(v9(5))
        f(v9(5)) = f(v9(6))
        f(v9(6)) = v9(7)
!
        do j =1,a9(i)-1
          v9(5) = v9(5) + c9(2)
          v9(6) = v9(4) + a9(j)
          v9(7) = f(v9(5))
          f(v9(5)) = f(v9(6))
          f(v9(6)) = v9(7)
        enddo
      end do
!
!***********************************************************************
!   stages 1 & 2
!   get two-element dhts
!***********************************************************************
!
9500  do i = 0,n-2,2
        t      = f(i) - f(i+1)
        f(i)   = f(i) + f(i+1)
        f(i+1) = t
      end do
!
!***********************************************************************
!   get four-element dhts
!***********************************************************************
!
      do i = 0,n-4,4
        t      = f(i) - f(i+2)
        f(i)   = f(i) + f(i+2)
        f(i+2) = t
        t      = f(i+1)-f(i+3)
        f(i+1) = f(i+1)+f(i+3)
        f(i+3) = t
      enddo

      if(p == 2) go to 9636  !  return
!
!***********************************************************************
!   stages 3, 4, ...
!***********************************************************************
!
      u9 = c9(6)
      sx9 = 4
!
      do l9 = 2,c9(6)
        v9(2) = sx9 + sx9
        u9 = u9-1
        v9(3) = m9(u9-1)

        do q9 = 0,c9(5),v9(2)
          i = q9
          d9 = i+sx9
          t     = f(i) - f(d9)
          f(i)  = f(i) + f(d9)
          f(d9) = t
          k9 = d9-1
!
          do j=v9(3),n9,v9(3)
            i  = i  + 1
            d9 = i  + sx9
            e9 = k9 + sx9
            t  = f(d9) + f(e9)*t9(j)
            x9 = f(e9) - t*s9(j)
            y9 = x9*t9(j) + t
            t = f(i) + y9
            f(d9) = f(i)  - y9
            f(e9) = f(k9) + x9
            f(k9) = f(k9) - x9
            f(i)  = t
            k9 = k9 -1
          enddo
        enddo
!
        sx9= v9(2)
!
      enddo
!
!***********************************************************************
!   save p so table generation can be skipped on next call
!***********************************************************************
!
9636  pold = p

   endsubroutine fhtfr 

end module hartley2d_module
